unit DescHint;
{
  Control that provide any interface for admin items requirint that the hint
  show the Caption, if this is hide.

  How to use:
  This control need to intercept the TApplication.OnShowHint event.  if you
  need to use this event, use TDescHint.OnShowHint instead.
  For specify the controls that must show the descriptive hint, in this control
  leave empty the hint property and put True int the ShowHint property.

  Limitation:
  Create only one TDescHint control for all the application.


  Control que provee una interfaz para administrar items que requieren que
  el hint muestre el Caption, si este est oculto.

  Cmo usarlo:
  Este control necesita interceptar el evento TApplication.OnShowHint, si usted
  desea usar este evento, use DescHint1.OnShowHint.
  Para especificar los controles que deben mostrar un hint descriptivo, en dicho
  control deje vaca la propiedad Hint y ponga True en la propiedad ShowHint

  Limitacin:
  Crear un solo control TDescHint para toda la aplicacin.

}
interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,
  Checklst,
  Grids, DBGrids, DB, DBTables;

type
  TCustomHintWindow = class;
  TDescHint = class(TComponent)
  private
    { Private declarations }
    FActive: Boolean;
    FClientHintRect: TRect;
    FDrawHintRect: TRect;
    FAlignment: TAlignment;
    FOnShowHint: TShowHintEvent;
    FFont: TFont;
  protected
    { Protected declarations }
    procedure DoShowHint(var HintStr: String;
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure GetHintInfoCustomListBox(
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure GetHintInfoCustomGrid(
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure GetHintInfoCustomEdit(
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure GetHintInfoCustomCheckBox(
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure GetHintInfoCustomLabel(
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure GetHintInfoControl(
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure SetFont(Value: TFont);
    property Font: TFont read FFont write SetFont;// stored IsFontStored;
    property Alignment: TAlignment read FAlignment write FAlignment;
    function CalcHintRect
      (MaxWidth: Integer; const AHint: string; AData: Pointer): TRect;
  public
    { Public declarations }
    procedure ShowHint(var HintStr: String;
      var CanShow: Boolean; var HintInfo: THintInfo);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property Active: Boolean read FActive write FActive default True;
    property OnShowHint: TShowHintEvent read FOnShowHint write FOnShowHint;
  end;
  TCustomHintWindow = class(THintWindow)
  private
//    FDescHint: TDescHint;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ActivateHint(Rect: TRect; const AHint: string); override;
  end;
var
  CDescHint: TDescHint;
  CHintWindow: TCustomHintWindow;
  CTwice: Integer;
implementation

type
  // Trampa para saltar el mecanismo de proteccin de propiedades y mtodos:
  // Sugerencia: no trampear funciones virtuales
  TPublicControl = class(TControl)
    public
    property Font;
    property Text;
  end;
  TPublicCustomGrid = class(TCustomGrid)
    public
    function CellRect(ACol, ARow: Longint): TRect;
    property Col;
    property Row;
    property ColCount;
    property RowCount;
  end;
  TPublicCustomDBGrid = class(TCustomDBGrid)
    public
    property Columns;
    property DataLink;
  end;
  TPublicCustomCheckBox = class(TCustomCheckBox)
  end;

function TPublicCustomGrid.CellRect(ACol, ARow: Longint): TRect;
begin
  Result := inherited CellRect(ACol, ARow);
end;
(*
function TPublicCustomGrid._GetEditText(ACol, ARow: Integer): string;
begin
  Result := GetEditText(Acol, ARow);
end;
*)
constructor TDescHint.Create;
begin
  inherited;
  Inc(CTWice);
  if CTWice > 1 then begin
//    Dec(CTwice);
    raise Exception.Create('Debe haber slo un DescHint por aplicacin');
  end;
  CDescHint := Self;
  FActive := True;
  FFont := TFont.Create;
  if not(csDesigning in ComponentState)then begin
    OnShowHint := Application.OnShowHint;
    Application.OnShowHint := ShowHint;
  end;
end;

destructor TDescHint.Destroy;
begin
  CDescHint := nil;
  if not (csDesigning in ComponentState)then begin
    Application.OnShowHint := OnShowHint;
  end;
  Dec(CTwice);
  FFont.Free;
  inherited;
end;

procedure TDescHint.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TDescHint.DoShowHint(var HintStr: String;
  var CanShow: Boolean; var HintInfo: THintInfo);
begin
  if Assigned(FOnShowHint) then
    FOnShowHint(HintStr, CanShow, HintInfo);
end;

procedure TDescHint.GetHintInfoCustomListBox(
    var CanShow: Boolean; var HintInfo: THintInfo);
var
  i: Integer;
begin
  with HintInfo do begin
    with TCustomListBox(HintControl) do begin
      i := ItemAtPos(HintInfo.CursorPos, true);
      if (i < 0)or(i >= Items.Count) then begin
        CanShow := False;
      end
      else begin
        HintStr := Items[i];
        FClientHintRect := ItemRect(i);
        with FClientHintRect.TopLeft do
          if HintControl is TCheckListBox then
          begin
            Inc(X, 14);
            Dec(Y, 3);
          end
          else begin
            Dec(X, 1);
            Dec(Y, 3);
          end;
        with FClientHintRect do begin
          TopLeft := ClientToScreen(TopLeft);
          BottomRight := ClientToScreen(BottomRight);
        end;
        FDrawHintRect := FClientHintRect;
      end;
    end;
  end;
end;

procedure TDescHint.GetHintInfoCustomGrid;
var
  G: TGridCoord;
  i: Integer;
begin
  with HintInfo do begin
    ReshowTimeout := 100;
    with TPublicCustomGrid(HintControl) do begin
      G := MouseCoord(CursorPos.x, CursorPos.y);
      FDrawHintRect := CellRect(G.X, G.Y);
      with FDrawHintRect do begin
        TopLeft := ClientToScreen(TopLeft);
        BottomRight := ClientToScreen(BottomRight);
        Dec(Left);
        Dec(Top);
      end;
      FClientHintRect.TopLeft := FDrawHintRect.TopLeft;
      if (G.X >= ColCount)or(G.X < 0)
          or(G.Y >= RowCount)or(G.Y < 0) then
        CanShow := False
      else begin
        with FClientHintRect do begin
          Right := Left + ColWidths[G.X] + GridLineWidth;
          Bottom := Top + RowHeights[G.Y] + GridLineWidth;
        end;
        if not(HintControl is TCustomDBGrid) then
          HintStr := TPublicCustomGrid(HintControl).GetEditText(G.X, G.Y)
        else begin
          with TPublicCustomDBGrid(HintControl) do begin
            if (G.X = 0) then
              CanShow := False
            else begin
              if G.Y = 0 then begin
                HintStr := Columns[G.X - 1].Title.Caption;
                Self.Font := Columns[G.X - 1].Title.Font;
                FAlignment := Columns[G.X -1].Title.Alignment;
              end
              else begin
                if not DataLink.Active then
                  CanShow := False
                else begin
                  i := DataLink.ActiveRecord;
                  try
                    DataLink.ActiveRecord := DataLink.ActiveRecord + G.Y - Row;
                    HintStr := Fields[G.X - 1].AsString;
                  finally
                    DataLink.ActiveRecord := i;
                  end;
                  Self.Font := Columns[G.X - 1].Font;
                  FAlignment := Columns[G.X - 1].Alignment;
                end;
              end;
            end;
          end;
        end
      end;
    end;
  end;
end;

procedure TDescHint.GetHintInfoCustomEdit;
begin
  Dec(FClientHintRect.Top, 2);
  Dec(FClientHintRect.Left, 2);
  FDrawHintRect := FClientHintRect;
end;

procedure TDescHint.GetHintInfoCustomCheckBox;
begin
  with TPublicCustomCheckBox(HintInfo.HintControl) do begin
    Dec(FClientHintRect.Top, 3);
    if Alignment = taLeftJustify then begin
      Dec(FClientHintRect.Left, 1);
    end
    else begin
      Inc(FClientHintRect.Left, 15);
    end;
  end;
  FDrawHintRect := FClientHintRect;
end;

procedure TDescHint.GetHintInfoCustomLabel;
begin
  Dec(FClientHintRect.Top, 3);
  Dec(FClientHintRect.Left, 3);
  FDrawHintRect := FClientHintRect;
end;

procedure TDescHint.GetHintInfoControl;
begin
  with HintInfo do begin
    Self.Font := TPublicControl(HintControl).Font;
    FAlignment := taLeftJustify;
    if HintControl is TCustomListBox then
      GetHintInfoCustomListBox(CanShow, HintInfo)
    else if HintInfo.HintControl is TCustomGrid then
      GetHintInfoCustomGrid(CanShow, HintInfo)
//    else if HintInfo.HintControl is TCustomLabel then
//      GetHintInfoCustomLabel(CanShow, HintInfo)
    else begin
      HintStr := TPublicControl(HintControl).Text;
      if not Assigned(HintControl.Parent) then
        CanShow := False
      else begin
        FClientHintRect.TopLeft := HintControl
          .ClientToScreen(HintControl.ClientRect.TopLeft);
        FClientHintRect.BottomRight := HintControl
          .ClientToScreen(HintControl.ClientRect.BottomRight);
        FDrawHintRect := FClientHintRect;
        if HintInfo.HintControl is TCustomEdit then
          GetHintInfoCustomEdit(CanShow, HintInfo)
        else if HintInfo.HintControl is TCustomLabel then
          GetHintInfoCustomLabel(CanShow, HintInfo)
        else if HintInfo.HintControl is TCustomCheckBox then
          GetHintInfoCustomCheckBox(CanShow, HintInfo)
      end;
    end;
  end;
end;

function TDescHint.CalcHintRect
    (MaxWidth: Integer; const AHint: string; AData: Pointer): TRect;
begin
  if Assigned(CHintWindow) then begin
    with CHintWindow do begin
      Canvas.Font := Self.Font;
      Result := CalcHintRect(MaxWidth, AHint, AData);
    end;
  end
  else begin
    with TCustomHintWindow.Create(Self) do begin
      try
        Canvas.Font := Self.Font;
        Result := CalcHintRect(MaxWidth, AHint, AData);
      finally
        Free;
      end;
    end;
  end;
end;

procedure TDescHint.ShowHint(var HintStr: String;
  var CanShow: Boolean; var HintInfo: THintInfo);
var
  R: TRect;
  function MaxInt(x, y: Integer): Integer;
  begin
    if x > y then
      Result := x
    else
      Result := y;
  end;
  function MinInt(x, y: Integer): Integer;
  begin
    if x < y then
      Result := x
    else
      Result := y;
  end;
begin
  DoShowHint(HintStr, CanShow, HintInfo);
  HintInfo.HintWindowClass := TCustomHintWindow;
  if HintStr <> '' then begin
    if Assigned(CHintWindow) then
      CHintWindow.Canvas.Font := CHintWindow.Font;
  end
  else begin
    try
      GetHintInfoControl(CanShow, HintInfo);
      if CanShow then begin
        R := CalcHintRect(Screen.Width, HintInfo.HintStr, nil);
        with FClientHintRect do begin
          HintInfo.HintPos.y := Top;
          if FAlignment = taLeftJustify then begin
            HintInfo.HintPos.x := FClientHintRect.Left;
          end
          else if FAlignment = taRightJustify then begin
            HintInfo.HintPos.x := R.Left - R.Right + Right;
          end
          else if FAlignment = taCenter then begin
            HintInfo.HintPos.x :=
              (R.Left - R.Right)div 2 + (1 + Right + Left)div 2;
          end;
        end;
        if (FDrawHintRect.Left <= HintInfo.HintPos.x)
            and(FDrawHintRect.Right >= R.Right - R.Left + HintInfo.HintPos.x) then
          CanShow := False
        else begin
          with HintInfo.HintPos do with Screen do with R do begin
            x := MaxInt(0, MinInt(x, Width - Right + Left));
            y := MaxInt(0, MinInt(y, Height- Bottom + Top));
          end;
        end;
      end;
    except
      on Exception do
        CanShow := False;
    end;
    if Assigned(CHintWindow) then begin
      if CanShow then
        CHintWindow.Canvas.Font := Self.Font
      else
        CHintWindow.ReleaseHandle;
    end;
  end;
end;

constructor TCustomHintWindow.Create;
begin
  inherited;
  CHintWindow := Self;
  Canvas.Font := CDescHint.Font;
end;

destructor TCustomHintWindow.Destroy;
begin
  CHintWindow := nil;
  inherited;
end;

procedure TCustomHintWindow.ActivateHint(Rect: TRect; const AHint: string);
begin
  inherited ActivateHint(Rect, AHint);
end;

initialization
  CTwice := 0;
end.

